#include <dirent.h>
#include <errno.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "arch.h"
#include "bool.h"
#include "env.h"
#include "extr.h"
#include "fs.h"
#include "link.h"
#include "math.h"
#include "msg.h"
#include "opts.h"
#include "path.h"
#include "str.h"



const unsigned int Tries = 5;
const char Mark = '%';



static BOOL setup(struct archive *arch, char *arg) {
	char *b;
	size_t ext, dot = 0, suf = 0, len;

	arch->file.p = arg;
	arch->file.e = ent_from_path(arg);
	arch->file.z = arch->file.e + strlen(arch->file.e);

	len = elen(arch->file);
	if (Extension) {
		ext = strlen(Extension);
		if (
			len < ext + 2 ||
			arch->file.z[-ext-1] != '.' ||
			strcasecmp(arch->file.z - ext, Extension) != 0
		) len += (dot = 1) + (suf = ext);
	}

	arch->links.file.e = b = emalloc(len + 1);
	mem_cat(&b, arch->file.e, elen(arch->file));
	mem_cat(&b, ".", dot);
	mem_cat(&b, Extension, suf);
	arch->links.file.z = b;
	*b = 0;

	if (!(arch->extr = find_extr(arch))) {
		msg(MSG_NO_MATCH_EXTR, NULL, arg);
		return(FALSE);
	}

	arch->pos = arch->extr->pos;
	str_to_lower(arch->links.file.e + arch->pos);
	return(TRUE);
}


static void mark_path(char *buf, const unsigned int try, const char *entry, const size_t len) {
	if (try > 0)
		buf[try - 1] = Mark;
	memcpy(buf + try, entry, len);
	buf[try + len] = 0;
}


static BOOL create_sandbox(struct archive *arch) {
	char *b;
	unsigned int i;
	size_t len = Absolute ? dlen(arch->file) : 0;

	arch->sandbox.p = b = emalloc(len + Tries - 1 + arch->pos);
	mem_cat(&b, arch->file.p, len);
	arch->sandbox.e = b;

	for (i = 0; i < Tries; ++i) {
		mark_path(b, i, arch->links.file.e, arch->pos - 1);
		if (mkdir(arch->sandbox.p, S_IRWXU | S_IRWXG | S_IRWXO) == 0) {
			arch->sandbox.z = arch->sandbox.e + i + arch->pos - 1;
			return TRUE;
		}
		if (errno != EEXIST)
			break;
	}
	err(MSG_MKDIR, arch->sandbox.p);
	free(arch->sandbox.p);
	arch->sandbox.p = NULL;
	return FALSE;
}


static BOOL extract(struct archive *arch) {
	int status;
	pid_t pid = fork();

	if (pid == 0) {
		if (chdir(arch->sandbox.p) != 0)
			die(MSG_CHDIR, arch->sandbox.p);

		if (dup2(2, 1) != 1)
			die(MSG_STDIN_TO_STDERR, NULL);

		if (!set_extr_env(arch))
			die(MSG_PUTENV, NULL);

		/* never pass a file starting with a dash to the extractor */
		if (*arch->links.file.e == '-')
			*(arch->links.file.e -= 2) = '.';

		execl(arch->extr->extract, "extract", arch->links.file.e, (char *)NULL);
		die(MSG_EXEC, arch->extr->extract);
	}

	waitpid(pid, &status, 0);
	if (WEXITSTATUS(status) == 0)
		return TRUE;
	msg(MSG_NOT_EXTRACT, NULL, arch->file.p);
	return FALSE;
}


static BOOL run_helper(char **cmd, char *m, char *id) {
	int status;
	pid_t pid = fork();

	if (pid == 0) {
		if (dup2(2, 1) != 1)
			die(MSG_STDIN_TO_STDERR, NULL);
		execvp(cmd[0], cmd);
		die(MSG_EXEC, cmd[0]);
	}
	waitpid(pid, &status, 0);
	if (WEXITSTATUS(status) == 0)
		return TRUE;
	msg(m, NULL, id);
	return FALSE;
}


static BOOL delete_partial(struct archive *arch) {
	char *cmd[] = { "rm", "-rf", "--", NULL, NULL };

	cmd[3] = arch->sandbox.p;
	return run_helper(cmd, MSG_REMOVE_PARTIAL, arch->sandbox.p);
}


static BOOL norm_perms(struct archive *arch) {
	char *cmd[] = { "chmod", "-R", "a=,+rwX", NULL, NULL };

	if (!arch->extr->perms)
		return TRUE;
	cmd[3] = arch->sandbox.p;
	return run_helper(cmd, MSG_NORM_PERMS, arch->sandbox.p);
}


static char *move_out(struct archive *arch, char *entry, size_t entry_len) {
	char *path1, *path2, *b;
	unsigned int i, j;
	int ret;
	BOOL success = FALSE;

	/*
	 *	rename parent/dir/entry to parent/%*entry
	 *	rmdir parent/dir
	 *	remove as many % characters as possible
	 */

	path1 = b = emalloc(dlen(arch->sandbox) + MAX(Tries - 1, elen(arch->sandbox) + 1) +
	                    entry_len + 1);
	mem_cat(&b, arch->sandbox.p, plen(arch->sandbox));
	*b++ = '/';
	mem_cat(&b, entry, entry_len + 1);

	path2 = emalloc(dlen(arch->sandbox) + Tries + entry_len + 1);
	memcpy(path2, arch->sandbox.p, dlen(arch->sandbox));

	/* For the temporary rename, use one try more to avoid failing to
	 * (in the end) rename parent/%%%%foo/%%%%foo to parent/%%%%foo */
	for (i = 0; i <= Tries; ++i) {
		mark_path(path2 + dlen(arch->sandbox), i, entry, entry_len);
		ret = safe_rename_v(path1, path2);
		if (ret == 1)
			success = TRUE;
		if (ret != 0)
			break;
	}

	if (i > Tries) {
		err(MSG_RENAME, path1);
		delete_partial(arch);
	} else if (rmdir(arch->sandbox.p) != 0)
		err(MSG_REMOVE, arch->sandbox.p);

	if (!success) {
		free(path1);
		free(path2);
		return NULL;
	}

	for (j = 0; i != j && j < Tries; ++j) {
		mark_path(path1 + dlen(arch->sandbox), j, entry, entry_len);
		if (safe_rename_v(path2, path1) == 1) {
			free(path2);
			return path1;
		}
	}

	/* could not lower the number of % characters */
	free(path1);
	return path2;
}


static BOOL handle_contents(struct archive *arch) {
	DIR *dir;
	struct dirent *ent;
	char *first = NULL, *moved;
	unsigned int entries = 0;
	size_t len = 0;
	BOOL success = TRUE;

	if (!(dir = opendir_v(arch->sandbox.p)))
		return FALSE;
	while ((ent = readdir(dir))) {
		if (is_dot(ent->d_name))
			continue;
		++entries;
		if (entries == 1) {
			len = strlen(ent->d_name);
			first = emalloc(len + 1);
			memcpy(first, ent->d_name, len + 1);
		} else if (entries == 2)
			break;
	}
	closedir(dir);

	if (entries == 0) {
		msg(MSG_EMPTY_ARCH, NULL, arch->file.p);
		if (rmdir(arch->sandbox.p) == 0)
			return TRUE;
		err(MSG_REMOVE, arch->sandbox.p);
		return FALSE;
	}

	if (entries == 1 && !arch->extr->subdir) {
		if ((moved = move_out(arch, first, len))) {
			pretty_print_ent(moved);
			free(moved);
		} else
			success = FALSE;
	} else
		pretty_print_ent(arch->sandbox.p);

	free(first);
	return success;
}


static BOOL remove_archive_files(struct archive *arch) {
	BOOL success = TRUE;
	char *path, *b;
	struct link *l;
	size_t longest = 0;

	if (!Unlink)
		return TRUE;

	for (l = &arch->links; l; l = l->next)
		longest = MAX(longest, elen(l->file));
	path = b = emalloc(dlen(arch->file) + longest + 1);
	mem_cat(&b, arch->file.p, dlen(arch->file));

	for (l = &arch->links; l; l = l->next) {
		strcpy(b, l->file.e);
		if (!unlink(path) == 0) {
			err(MSG_REMOVE, path);
			success = FALSE;
		}
	}

	free(path);
	return success;
}


static void cleanup(struct archive *arch) {
	if (arch->sandbox.p)
		free(arch->sandbox.p);
	free_links(arch);
}


static BOOL verify_file(char *path) {
	struct stat st;
	int ret = stat(path, &st);

	if (ret != 0) {
		if (errno == ENOENT)
			msg(MSG_NOT_EXIST, NULL, path);
		else
			err(MSG_STAT, path);
		return FALSE;
	}

	if (!S_ISDIR(st.st_mode))
		return TRUE;
	msg(MSG_IS_DIR, NULL, path);
	return FALSE;
}


BOOL generic_extract(char *arg) {
	struct archive arch = { 0 };
	BOOL success;

	if (!verify_file(arg))
		return FALSE;

	if (!setup(&arch, arg) || !create_sandbox(&arch) || !build_links(&arch)) {
		cleanup(&arch);
		return FALSE;
	}

	success = extract(&arch);

	if (!success && !Keep) {
		delete_partial(&arch);
		cleanup(&arch);
		return FALSE;
	}

	success = remove_symlinks(&arch) && success;
	success = norm_perms(&arch)      && success;
	success = handle_contents(&arch) && success;
	success = success && remove_archive_files(&arch);

	cleanup(&arch);
	return success;
}
