#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "arch.h"
#include "bool.h"
#include "extr.h"
#include "fs.h"
#include "link.h"
#include "math.h"
#include "msg.h"
#include "opts.h"
#include "path.h"
#include "str.h"



BOOL make_symlink(struct link *l) {
	ino_t inode;

	if (symlink(l->target.p, l->file.p) != 0) {
		err(MSG_SYMLINK, l->file.p);
		return FALSE;
	}

	if (!get_inode(&inode, l->file.p)) {
		err(MSG_STAT, l->file.p);
		return FALSE;
	}

	l->inode = inode;
	return TRUE;
}


void complete_first_link(struct archive *arch) {
	char *b;
	size_t dot, dir;

	/* We are finally in possession of the extraction directory name
	 * needed to build the link path. */
	arch->links.file.p = b = emalloc(plen(arch->sandbox) + 1 + elen(arch->links.file) + 1);
	mem_cat(&b, arch->sandbox.p, plen(arch->sandbox));
	*b++ = '/';
	strcpy(b, arch->links.file.e);

	/* replace the freestanding entry string with a pointer into the path */
	arch->links.file.z = b + elen(arch->links.file);
	free(arch->links.file.e);
	arch->links.file.e = b;

	/* build the archive link target */
	dot = (Absolute || *arch->file.p != '/') ? 3 : 0;
	dir = Absolute ? 0 : dlen(arch->file);
	arch->links.target.p = b = emalloc(dot + dir + elen(arch->file) + 1);
	mem_cat(&b, "../", dot);
	mem_cat(&b, arch->file.p, dir);
	arch->links.target.e = b;
	mem_cat(&b, arch->file.e, elen(arch->file));
	arch->links.target.z = b;
	*b = 0;
}


BOOL build_links(struct archive *arch) {
	char *slash, *b;
	DIR *dir;
	struct dirent *ent;
	size_t ent_len;
	struct link *old;
	BOOL success = TRUE;

	/* first link is peculiar */
	complete_first_link(arch);
	if (!make_symlink(&arch->links))
		return FALSE;

	/* open the directory the archive resides in */
	if (dlen(arch->file) == 0)
		dir = opendir_v(".");
	else {
		slash = arch->file.e - 1;
		*slash = 0;
		dir = opendir_v(arch->file.p);
		*slash = '/';
	}
	if (!dir)
		return FALSE;

	while ((ent = readdir(dir))) {
		/* main name different from first link's? */
		if (strncmp(arch->links.file.e, ent->d_name, arch->pos) != 0)
			continue;
		/* extension equal to first link's? */
		if (strcasecmp(arch->links.file.e + arch->pos, ent->d_name + arch->pos) == 0)
			continue;
		match_extr(arch->extr, ent->d_name);
		/* extension didn't match or at position different from first link's? */
		if (arch->pos != arch->extr->pos)
			continue;

		ent_len = strlen(ent->d_name);

		old = arch->links.next;
		arch->links.next = emalloc(sizeof(struct link));
		arch->links.next->next = old;

		/* build link target */
		arch->links.next->target.p = b = emalloc(dlen(arch->links.target) + ent_len + 1);
		mem_cat(&b, arch->links.target.p, dlen(arch->links.target));
		mem_cat(&b, ent->d_name, ent_len);
		arch->links.next->target.z = b;
		*b = 0;

		/* build link path and link file */
		arch->links.next->file.p = b = emalloc(plen(arch->sandbox) + 1 + ent_len + 1);
		mem_cat(&b, arch->links.file.p, plen(arch->sandbox) + 1);
		arch->links.next->file.e = b;
		arch->links.next->file.z = b + ent_len;
		mem_cat(&b, ent->d_name, arch->extr->pos);
		str_cpy_lower(b, ent->d_name + arch->extr->pos);

		/* create the link */
		if (!make_symlink(arch->links.next)) {
			success = FALSE;
			break;
		}
	}

	closedir(dir);
	return success;
}


BOOL remove_symlink(struct link *l) {
	ino_t inode;
	char *target;
	ssize_t len;
	BOOL success = TRUE;

	if (!get_inode(&inode, l->file.p)) {
		if (errno == ENOENT)  /* extractor deleted link */
			return TRUE;
		else {
			err(MSG_STAT, l->file.p);
			return FALSE;
		}
	}

	if (l->inode != inode)  /* extractor replaced link */
		return TRUE;

	target = emalloc(plen(l->target) + 1);
	len = readlink(l->file.p, target, plen(l->target));
	if (len < 0) {
		err(MSG_READLINK, l->file.p);
		success = FALSE;
	} else if (len == plen(l->target) && memcmp(target, l->target.p, len) == 0) {
		/* We know the link has its original inode number and target.
		 * This is probably the link we created; let's delete it. */
		if (unlink(l->file.p) != 0 && errno != ENOENT) {
			err(MSG_REMOVE, l->file.p);
			success = FALSE;
		}
	}

	free(target);
	return success;
}


BOOL remove_symlinks(struct archive *arch) {
	struct link *l;
	BOOL success = TRUE;

	if (arch->links.file.p)
		success = remove_symlink(&arch->links);
	for (l = arch->links.next; l; l = l->next)
		success = remove_symlink(l) && success;
	return success;
}


void free_links(struct archive *arch) {
	struct link *l, *next;
	if (arch->links.file.p) {
		free(arch->links.file.p);
		free(arch->links.target.p);
	} else
		free(arch->links.file.e);

	for (l = arch->links.next; l; l = next) {
		free(l->file.p);
		free(l->target.p);
		next = l->next;
		free(l);
	}
}
